"""BEMServer API client resources

/about/ endpoint
"""
from .base import BaseResources


class AboutResources(BaseResources):
    endpoint_base_uri = "/about/"
    disabled_endpoints = ["getone", "create", "update", "delete"]



"""BEMServer API client resources

/analysis/ endpoints
"""
from .base import BaseResources
from ..enums import BucketWidthUnit, StructuralElement
from ..exceptions import BEMServerAPIClientValueError


class AnalysisResources(BaseResources):
    endpoint_base_uri = "/analysis/"
    disabled_endpoints = ["getall", "getone", "create", "update", "delete"]

    def get_completeness(
        self,
        start_time,
        end_time,
        timeseries,
        data_state,
        bucket_width_value,
        bucket_width_unit,
        timezone="UTC",
        *,
        etag=None,
    ):
        if bucket_width_unit not in BucketWidthUnit:
            raise BEMServerAPIClientValueError(
                f"Invalid bucket width unit: {bucket_width_unit}"
            )

        endpoint = f"{self.endpoint_base_uri}completeness"
        q_params = {
            "start_time": start_time,
            "end_time": end_time,
            "timeseries": timeseries,
            "data_state": data_state,
            "bucket_width_value": bucket_width_value,
            "bucket_width_unit": bucket_width_unit.value,
            "timezone": timezone,
        }
        return self._req.getall(endpoint, etag=etag, params=q_params)

    def get_energy_consumption_breakdown(
        self,
        structural_element_type,
        structural_element_id,
        start_time,
        end_time,
        bucket_width_value,
        bucket_width_unit,
        timezone="UTC",
        *,
        etag=None,
    ):
        if structural_element_type not in StructuralElement:
            raise BEMServerAPIClientValueError(
                f"Invalid structural element type: {structural_element_type}"
            )
        if bucket_width_unit not in BucketWidthUnit:
            raise BEMServerAPIClientValueError(
                f"Invalid bucket width unit: {bucket_width_unit}"
            )

        endpoint = (
            f"{self.endpoint_base_uri}energy_consumption/"
            f"{structural_element_type.value}/{structural_element_id}"
        )
        q_params = {
            "start_time": start_time,
            "end_time": end_time,
            "bucket_width_value": bucket_width_value,
            "bucket_width_unit": bucket_width_unit.value,
            "timezone": timezone,
        }
        return self._req.getall(endpoint, etag=etag, params=q_params)



"""BEMServer API client services resources"""

from .cleanup import (  # noqa
    ST_CleanupByCampaignResources,
    ST_CleanupByTimeseriesResources,
)
from .check_missing import (  # noqa
    ST_CheckMissingByCampaignResources,
)
from .check_outliers import (  # noqa
    ST_CheckOutlierByCampaignResources,
)



"""BEMServer API client resources

/timeseries/ endpoints
/timeseries_data_states/ endpoints
/timeseries_properties/ endpoints
/timeseries_property_data/ endpoints
/timeseries_data/ endpoints
/timeseries_by_sites/ endpoints
/timeseries_by_buildings/ endpoints
/timeseries_by_storeys/ endpoints
/timeseries_by_spaces/ endpoints
/timeseries_by_zones/ endpoints
/timeseries_by_events/ endpoints
"""
from .base import BaseResources
from ..enums import DataFormat, Aggregation, BucketWidthUnit


class TimeseriesResources(BaseResources):
    endpoint_base_uri = "/timeseries/"


class TimeseriesDataStateResources(BaseResources):
    endpoint_base_uri = "/timeseries_data_states/"


class TimeseriesPropertyResources(BaseResources):
    endpoint_base_uri = "/timeseries_properties/"


class TimeseriesPropertyDataResources(BaseResources):
    endpoint_base_uri = "/timeseries_property_data/"


class TimeseriesDataResources(BaseResources):
    endpoint_base_uri = "/timeseries_data/"
    disabled_endpoints = ["getall", "getone", "create", "update"]

    def endpoint_uri_by_campaign(self, campaign_id):
        return f"{self.endpoint_base_uri}campaign/{str(campaign_id)}/"

    def upload(self, data_state, data, format=DataFormat.json):
        """

        :param int data_state: timeseries data state id to feed
        :param str data: data to upload (for example a read content of file stream)
        :param DataFormat format: (optional, default JSON)
            data format, either CSV or JSON
        """
        return self._req.upload_data(
            self.endpoint_base_uri,
            data,
            format=format,
            params={"data_state": data_state},
        )

    def upload_by_names(self, campaign_id, data_state, data, format=DataFormat.json):
        """

        :param int data_state: timeseries data state id to feed
        :param str data: data to upload (for example a read content of file stream)
        :param DataFormat format: (optional, default JSON)
            data format, either CSV or JSON
        """
        return self._req.upload_data(
            self.endpoint_uri_by_campaign(campaign_id),
            data,
            format=format,
            params={"data_state": data_state},
        )

    def download(
        self,
        start_time,
        end_time,
        data_state,
        timeseries_ids,
        timezone="UTC",
        format=DataFormat.json,
    ):
        return self._req.download(
            self.endpoint_base_uri,
            format=format,
            params={
                "start_time": start_time,
                "end_time": end_time,
                "data_state": data_state,
                "timeseries": timeseries_ids,
                "timezone": timezone,
            },
        )

    def download_by_names(
        self,
        campaign_id,
        start_time,
        end_time,
        data_state,
        timeseries_names,
        timezone="UTC",
        format=DataFormat.json,
    ):
        return self._req.download(
            self.endpoint_uri_by_campaign(campaign_id),
            format=format,
            params={
                "start_time": start_time,
                "end_time": end_time,
                "data_state": data_state,
                "timeseries": timeseries_names,
                "timezone": timezone,
            },
        )

    def download_aggregate(
        self,
        start_time,
        end_time,
        data_state,
        timeseries_ids,
        timezone="UTC",
        aggregation=Aggregation.avg,
        bucket_width_value="1",
        bucket_width_unit=BucketWidthUnit.hour,
        format=DataFormat.json,
    ):
        return self._req.download(
            f"{self.endpoint_base_uri}aggregate",
            format=format,
            params={
                "start_time": start_time,
                "end_time": end_time,
                "data_state": data_state,
                "timeseries": timeseries_ids,
                "timezone": timezone,
                "aggregation": aggregation.value,
                "bucket_width_value": bucket_width_value,
                "bucket_width_unit": bucket_width_unit.value,
            },
        )

    def download_aggregate_by_names(
        self,
        campaign_id,
        start_time,
        end_time,
        data_state,
        timeseries_names,
        timezone="UTC",
        aggregation=Aggregation.avg,
        bucket_width_value="1",
        bucket_width_unit=BucketWidthUnit.hour,
        format=DataFormat.json,
    ):
        return self._req.download(
            f"{self.endpoint_uri_by_campaign(campaign_id)}aggregate",
            format=format,
            params={
                "start_time": start_time,
                "end_time": end_time,
                "data_state": data_state,
                "timeseries": timeseries_names,
                "timezone": timezone,
                "aggregation": aggregation.value,
                "bucket_width_value": bucket_width_value,
                "bucket_width_unit": bucket_width_unit.value,
            },
        )

    def delete(self, start_time, end_time, data_state, timeseries_ids):
        return self._req._execute(
            "DELETE",
            self.endpoint_base_uri,
            params={
                "start_time": start_time,
                "end_time": end_time,
                "data_state": data_state,
                "timeseries": timeseries_ids,
            },
        )

    def delete_by_names(
        self,
        campaign_id,
        start_time,
        end_time,
        data_state,
        timeseries_names,
    ):
        return self._req._execute(
            "DELETE",
            self.endpoint_uri_by_campaign(campaign_id),
            params={
                "start_time": start_time,
                "end_time": end_time,
                "data_state": data_state,
                "timeseries": timeseries_names,
            },
        )


class TimeseriesBySiteResources(BaseResources):
    endpoint_base_uri = "/timeseries_by_sites/"
    disabled_endpoints = ["update"]


class TimeseriesByBuildingResources(BaseResources):
    endpoint_base_uri = "/timeseries_by_buildings/"
    disabled_endpoints = ["update"]


class TimeseriesByStoreyResources(BaseResources):
    endpoint_base_uri = "/timeseries_by_storeys/"
    disabled_endpoints = ["update"]


class TimeseriesBySpaceResources(BaseResources):
    endpoint_base_uri = "/timeseries_by_spaces/"
    disabled_endpoints = ["update"]


class TimeseriesByZoneResources(BaseResources):
    endpoint_base_uri = "/timeseries_by_zones/"
    disabled_endpoints = ["update"]


class TimeseriesByEventResources(BaseResources):
    endpoint_base_uri = "/timeseries_by_events/"
    disabled_endpoints = ["update"]



"""BEMServer API client resources

/io/ endpoints
"""
from .base import BaseResources


class IOResources(BaseResources):
    endpoint_base_uri = "/io/"
    disabled_endpoints = ["getall", "getone", "create", "update", "delete"]

    def upload_timeseries_csv(self, campaign_id, csv_files):
        """

        :param dict csv_files:
            key is the upload field name (timeseries_csv)
            value is a file stream (tempfile.SpooledTemporaryFile)
        """
        endpoint = f"{self.endpoint_base_uri}timeseries"
        q_params = {"campaign_id": campaign_id}
        return self._req.upload_files(endpoint, params=q_params, files=csv_files)

    def upload_sites_csv(self, campaign_id, csv_files):
        """

        :param dict csv_files:
            key is the upload field name (sites_csv, buildings_csv...)
            value is a file stream (tempfile.SpooledTemporaryFile)
        """
        endpoint = f"{self.endpoint_base_uri}sites"
        q_params = {"campaign_id": campaign_id}
        return self._req.upload_files(endpoint, params=q_params, files=csv_files)



"""BEMServer API client cleanup service resources

/st_cleanups_by_campaigns/ endpoints
/st_cleanups_by_timeseries/ endpoints
"""
from ..base import BaseResources


class ST_CleanupByCampaignResources(BaseResources):
    endpoint_base_uri = "/st_cleanups_by_campaigns/"

    def get_full(self, *, etag=None, **kwargs):
        endpoint = f"{self.endpoint_base_uri}full"
        return self._req.getall(endpoint, etag=etag, params=kwargs)


class ST_CleanupByTimeseriesResources(BaseResources):
    endpoint_base_uri = "/st_cleanups_by_timeseries/"
    disabled_endpoints = ["create", "update", "delete"]

    def get_full(self, *, etag=None, **kwargs):
        endpoint = f"{self.endpoint_base_uri}full"
        return self._req.getall(endpoint, etag=etag, params=kwargs)



"""BEMServer API client"""
import logging
from packaging.version import Version, InvalidVersion
from requests.auth import HTTPBasicAuth

from .exceptions import BEMServerAPIVersionError
from .request import BEMServerApiClientRequest
from .resources import (
    AboutResources,
    UserResources,
    UserGroupResources,
    UserByUserGroupResources,
    CampaignResources,
    UserGroupByCampaignResources,
    CampaignScopeResources,
    UserGroupByCampaignScopeResources,
    TimeseriesResources,
    TimeseriesDataStateResources,
    TimeseriesPropertyResources,
    TimeseriesPropertyDataResources,
    TimeseriesDataResources,
    StructuralElementPropertyResources,
    SiteResources,
    SitePropertyResources,
    SitePropertyDataResources,
    BuildingResources,
    BuildingPropertyResources,
    BuildingPropertyDataResources,
    StoreyResources,
    StoreyPropertyResources,
    StoreyPropertyDataResources,
    SpaceResources,
    SpacePropertyResources,
    SpacePropertyDataResources,
    ZoneResources,
    ZonePropertyResources,
    ZonePropertyDataResources,
    TimeseriesBySiteResources,
    TimeseriesByBuildingResources,
    TimeseriesByStoreyResources,
    TimeseriesBySpaceResources,
    TimeseriesByZoneResources,
    IOResources,
    AnalysisResources,
    ST_CleanupByCampaignResources,
    ST_CleanupByTimeseriesResources,
    ST_CheckMissingByCampaignResources,
    ST_CheckOutlierByCampaignResources,
    EnergySourceResources,
    EnergyEndUseResources,
    EnergyConsumptionTimseriesBySiteResources,
    EnergyConsumptionTimseriesByBuildingResources,
    EventResources,
    EventCategoryResources,
    EventCategoryByUserResources,
    EventBySiteResources,
    EventByBuildingResources,
    EventByStoreyResources,
    EventBySpaceResources,
    EventByZoneResources,
    TimeseriesByEventResources,
    NotificationResources,
)


APICLI_LOGGER = logging.getLogger(__name__)

REQUIRED_API_VERSION = {
    "min": Version("0.11.1"),
    "max": Version("0.12.0"),
}


class BEMServerApiClient:
    """API client"""

    def __init__(
        self,
        host,
        use_ssl=True,
        authentication_method=None,
        uri_prefix="http",
        auto_check=False,
        request_manager=None,
    ):
        self.base_uri_prefix = uri_prefix or "http"
        self.host = host
        self.use_ssl = use_ssl

        self._request_manager = request_manager or BEMServerApiClientRequest(
            self.base_uri,
            authentication_method,
            logger=APICLI_LOGGER,
        )

        self.about = AboutResources(self._request_manager)

        if auto_check:
            api_version = self.about.getall().data["versions"]["bemserver_api"]
            self.check_api_version(api_version)

        self.io = IOResources(self._request_manager)

        self.users = UserResources(self._request_manager)
        self.user_groups = UserGroupResources(self._request_manager)
        self.user_by_user_groups = UserByUserGroupResources(self._request_manager)

        self.campaigns = CampaignResources(self._request_manager)
        self.user_groups_by_campaigns = UserGroupByCampaignResources(
            self._request_manager
        )
        self.campaign_scopes = CampaignScopeResources(self._request_manager)
        self.user_groups_by_campaign_scopes = UserGroupByCampaignScopeResources(
            self._request_manager
        )

        self.timeseries = TimeseriesResources(self._request_manager)
        self.timeseries_datastates = TimeseriesDataStateResources(self._request_manager)
        self.timeseries_properties = TimeseriesPropertyResources(self._request_manager)
        self.timeseries_property_data = TimeseriesPropertyDataResources(
            self._request_manager
        )
        self.timeseries_data = TimeseriesDataResources(self._request_manager)

        self.sites = SiteResources(self._request_manager)
        self.buildings = BuildingResources(self._request_manager)
        self.storeys = StoreyResources(self._request_manager)
        self.spaces = SpaceResources(self._request_manager)
        self.zones = ZoneResources(self._request_manager)

        self.structural_element_properties = StructuralElementPropertyResources(
            self._request_manager
        )
        self.site_properties = SitePropertyResources(self._request_manager)
        self.building_properties = BuildingPropertyResources(self._request_manager)
        self.storey_properties = StoreyPropertyResources(self._request_manager)
        self.space_properties = SpacePropertyResources(self._request_manager)
        self.zone_properties = ZonePropertyResources(self._request_manager)

        self.site_property_data = SitePropertyDataResources(self._request_manager)
        self.building_property_data = BuildingPropertyDataResources(
            self._request_manager
        )
        self.storey_property_data = StoreyPropertyDataResources(self._request_manager)
        self.space_property_data = SpacePropertyDataResources(self._request_manager)
        self.zone_property_data = ZonePropertyDataResources(self._request_manager)

        self.timeseries_by_sites = TimeseriesBySiteResources(self._request_manager)
        self.timeseries_by_buildings = TimeseriesByBuildingResources(
            self._request_manager
        )
        self.timeseries_by_storeys = TimeseriesByStoreyResources(self._request_manager)
        self.timeseries_by_spaces = TimeseriesBySpaceResources(self._request_manager)
        self.timeseries_by_zones = TimeseriesByZoneResources(self._request_manager)

        self.analysis = AnalysisResources(self._request_manager)

        self.st_cleanup_by_campaign = ST_CleanupByCampaignResources(
            self._request_manager
        )
        self.st_cleanup_by_timeseries = ST_CleanupByTimeseriesResources(
            self._request_manager
        )

        self.st_check_missing_by_campaign = ST_CheckMissingByCampaignResources(
            self._request_manager
        )

        self.st_check_outlier_by_campaign = ST_CheckOutlierByCampaignResources(
            self._request_manager
        )

        self.energy_sources = EnergySourceResources(self._request_manager)
        self.energy_end_uses = EnergyEndUseResources(self._request_manager)
        self.energy_cons_ts_by_sites = EnergyConsumptionTimseriesBySiteResources(
            self._request_manager
        )
        self.energy_cons_ts_by_buildings = (
            EnergyConsumptionTimseriesByBuildingResources(self._request_manager)
        )

        self.events = EventResources(self._request_manager)
        self.event_categories = EventCategoryResources(self._request_manager)
        self.event_categories_by_users = EventCategoryByUserResources(
            self._request_manager
        )
        self.event_by_sites = EventBySiteResources(self._request_manager)
        self.event_by_buildings = EventByBuildingResources(self._request_manager)
        self.event_by_storeys = EventByStoreyResources(self._request_manager)
        self.event_by_spaces = EventBySpaceResources(self._request_manager)
        self.event_by_zones = EventByZoneResources(self._request_manager)
        self.timeseries_by_events = TimeseriesByEventResources(self._request_manager)

        self.notifications = NotificationResources(self._request_manager)

    @property
    def uri_prefix(self):
        uri_prefix = self.base_uri_prefix
        if self.use_ssl:
            uri_prefix = self.base_uri_prefix.replace("http", "https")
        return f"{uri_prefix}://"

    @property
    def base_uri(self):
        return f"{self.uri_prefix}{self.host}"

    @staticmethod
    def make_http_basic_auth(email, password):
        return HTTPBasicAuth(
            email.encode(encoding="utf-8"),
            password.encode(encoding="utf-8"),
        )

    @classmethod
    def check_api_version(cls, api_version):
        try:
            version_api = Version(str(api_version))
        except InvalidVersion as exc:
            raise BEMServerAPIVersionError(f"Invalid API version: {str(exc)}")
        version_min = REQUIRED_API_VERSION["min"]
        version_max = REQUIRED_API_VERSION["max"]
        if not (version_min <= version_api < version_max):
            raise BEMServerAPIVersionError(
                f"API version ({str(version_api)}) not supported!"
                f" (expected: >={str(version_min)},<{str(version_max)})"
            )



"""BEMServer API client resources

/energy_sources/ endpoints
/energy_end_uses/ endpoints
/energy_consumption_timeseries_by_sites/ endpoints
/energy_consumption_timeseries_by_buildings/ endpoints
"""
from .base import BaseResources


class EnergySourceResources(BaseResources):
    endpoint_base_uri = "/energy_sources/"
    disabled_endpoints = ["getone", "create", "update", "delete"]


class EnergyEndUseResources(BaseResources):
    endpoint_base_uri = "/energy_end_uses/"
    disabled_endpoints = ["getone", "create", "update", "delete"]


class EnergyConsumptionTimseriesBySiteResources(BaseResources):
    endpoint_base_uri = "/energy_consumption_timeseries_by_sites/"


class EnergyConsumptionTimseriesByBuildingResources(BaseResources):
    endpoint_base_uri = "/energy_consumption_timeseries_by_buildings/"



"""BEMServer API client resources"""

from .about import AboutResources  # noqa
from .analysis import AnalysisResources  # noqa
from .campaigns import (  # noqa
    CampaignResources,
    UserGroupByCampaignResources,
    CampaignScopeResources,
    UserGroupByCampaignScopeResources,
)
from .energy import (  # noqa
    EnergySourceResources,
    EnergyEndUseResources,
    EnergyConsumptionTimseriesBySiteResources,
    EnergyConsumptionTimseriesByBuildingResources,
)
from .events import (  # noqa
    EventResources,
    EventCategoryResources,
    EventCategoryByUserResources,
    EventBySiteResources,
    EventByBuildingResources,
    EventByStoreyResources,
    EventBySpaceResources,
    EventByZoneResources,
)
from .io import IOResources  # noqa
from .notifications import NotificationResources  # noqa
from .services import (  # noqa
    ST_CleanupByCampaignResources,
    ST_CleanupByTimeseriesResources,
    ST_CheckMissingByCampaignResources,
    ST_CheckOutlierByCampaignResources,
)
from .structural_elements import (  # noqa
    SiteResources,
    BuildingResources,
    StoreyResources,
    SpaceResources,
    ZoneResources,
    StructuralElementPropertyResources,
    SitePropertyResources,
    BuildingPropertyResources,
    StoreyPropertyResources,
    SpacePropertyResources,
    ZonePropertyResources,
    SitePropertyDataResources,
    BuildingPropertyDataResources,
    StoreyPropertyDataResources,
    SpacePropertyDataResources,
    ZonePropertyDataResources,
)
from .timeseries import (  # noqa
    TimeseriesResources,
    TimeseriesDataStateResources,
    TimeseriesPropertyResources,
    TimeseriesPropertyDataResources,
    TimeseriesDataResources,
    TimeseriesBySiteResources,
    TimeseriesByBuildingResources,
    TimeseriesByStoreyResources,
    TimeseriesBySpaceResources,
    TimeseriesByZoneResources,
    TimeseriesByEventResources,
)
from .users import (  # noqa
    UserResources,
    UserGroupResources,
    UserByUserGroupResources,
)



"""BEMServer API client request"""
import logging
import os
import requests
import requests.exceptions as req_exc

from .enums import DataFormat
from .response import BEMServerApiClientResponse
from .exceptions import BEMServerAPIInternalError, BEMServerAPIClientValueError


class BEMServerApiClientRequest:
    """API client requester"""

    _ETAG_HEADER_BY_HTTP_METHOD = {
        "GET": "If-None-Match",
        "PUT": "If-Match",
        "DELETE": "If-Match",
    }

    def __init__(self, base_uri, authentication_method, *, logger=None):
        self.base_uri = base_uri

        self._logger = logger or logging.getLogger(__name__)
        self._session = requests.Session()
        self._session.auth = authentication_method

    def _build_uri(self, endpoint_uri):
        return f"{self.base_uri}{endpoint_uri}"

    def _prepare_accept_header(self, format):
        """

        :raises TypeError: when format is None or not an enum type
        :raises BEMServerAPIClientValueError: when format is not in DataFormat enum
        """
        if format not in DataFormat:
            raise BEMServerAPIClientValueError(f"Invalid data format: {format}")
        return {"Accept": format.value}

    def _prepare_etag_header(self, http_method, etag):
        etag_header = {}
        if etag is not None:
            try:
                etag_header = {
                    self._ETAG_HEADER_BY_HTTP_METHOD[http_method.upper()]: etag
                }
            except KeyError:
                pass
        return etag_header

    def _execute(self, http_method, endpoint, *, etag=None, **kwargs):
        full_endpoint_uri = self._build_uri(endpoint)
        headers = {
            **kwargs.pop("headers", {}),
            **self._prepare_etag_header(http_method, etag),
        }
        self._logger.debug(f"{http_method} {full_endpoint_uri}")
        try:
            raw_resp = self._session.request(
                http_method, full_endpoint_uri, headers=headers, **kwargs
            )
        except req_exc.RequestException as exc:
            self._logger.error(
                f"Unexpected error while requesting {full_endpoint_uri}: {exc}"
            )
            raise BEMServerAPIInternalError from exc
        return BEMServerApiClientResponse(raw_resp)

    @staticmethod
    def _exclude_empty_files(files):
        """

        :param dict files:
            key is the upload field name (csv_file, timeseries_csv or sites_csv...)
            value is a file stream (tempfile.SpooledTemporaryFile)
        """
        not_empty_files = {}
        for k, v in files.items():
            if v is not None and v.seek(0, os.SEEK_END) > 0:
                v.seek(0, os.SEEK_SET)
                not_empty_files[k] = v
        return not_empty_files

    def getall(self, endpoint, *, etag=None, **kwargs):
        return self._execute("GET", endpoint, etag=etag, **kwargs)

    def getone(self, endpoint, *, etag=None):
        return self._execute("GET", endpoint, etag=etag)

    def create(self, endpoint, payload):
        return self._execute("POST", endpoint, json=payload)

    def update(self, endpoint, payload, etag):
        return self._execute("PUT", endpoint, json=payload, etag=etag)

    def delete(self, endpoint, etag):
        return self._execute("DELETE", endpoint, etag=etag)

    def upload_files(self, endpoint, files, **kwargs):
        """Upload files.

        :param dict files:
            key is the upload field name (csv_file, timeseries_csv or sites_csv...)
            value is a file stream (tempfile.SpooledTemporaryFile)
        """
        not_empty_files = self._exclude_empty_files(files)
        return self._execute("POST", endpoint, files=not_empty_files, **kwargs)

    def upload_data(self, endpoint, data, *, format=DataFormat.json, **kwargs):
        """Upload data from specified format.

        :param str data: data to upload (for example a read content of file stream)
        :param DataFormat format: (optional, default JSON)
            data format, either CSV or JSON
        """
        kwargs["headers"] = {
            **kwargs.pop("headers", {}),
            **self._prepare_accept_header(format),
        }
        return self._execute("POST", endpoint, data=data, **kwargs)

    def download(self, endpoint, *, format=DataFormat.json, **kwargs):
        """Download data in specified format.

        :param DataFormat format: (optional, default JSON)
            data format, either CSV or JSON
        """
        kwargs["headers"] = {
            **kwargs.pop("headers", {}),
            **self._prepare_accept_header(format),
        }
        return self._execute("GET", endpoint, **kwargs)



"""BEMServer API client resources

/users/ endpoints
/user_groups/ endpoints
/users_by_user_groups/ endpoints
"""
from .base import BaseResources


class UserResources(BaseResources):
    endpoint_base_uri = "/users/"

    def set_admin(self, id, state, *, etag=None):
        endpoint = f"{self.enpoint_uri_by_id(id)}/set_admin"
        return self._req.update(endpoint, {"value": state}, etag=etag)

    def set_active(self, id, state, *, etag=None):
        endpoint = f"{self.enpoint_uri_by_id(id)}/set_active"
        return self._req.update(endpoint, {"value": state}, etag=etag)


class UserGroupResources(BaseResources):
    endpoint_base_uri = "/user_groups/"


class UserByUserGroupResources(BaseResources):
    endpoint_base_uri = "/users_by_user_groups/"
    disabled_endpoints = ["update"]



"""BEMServer API client resources

/sites/ endpoints
/buildings/ endpoints
/storeys/ endpoints
/spaces/ endpoints
/zones/ endpoints
/structural_element_properties/ endpoints
/site_properties/ endpoints
/building_properties/ endpoints
/storey_properties/ endpoints
/space_properties/ endpoints
/zone_properties/ endpoints
/site_property_data/ endpoints
/building_property_data/ endpoints
/storey_property_data/ endpoints
/space_property_data/ endpoints
/zone_property_data/ endpoints
"""
from .base import BaseResources


class SiteResources(BaseResources):
    endpoint_base_uri = "/sites/"


class BuildingResources(BaseResources):
    endpoint_base_uri = "/buildings/"


class StoreyResources(BaseResources):
    endpoint_base_uri = "/storeys/"


class SpaceResources(BaseResources):
    endpoint_base_uri = "/spaces/"


class ZoneResources(BaseResources):
    endpoint_base_uri = "/zones/"


class StructuralElementPropertyResources(BaseResources):
    endpoint_base_uri = "/structural_element_properties/"


class SitePropertyResources(BaseResources):
    endpoint_base_uri = "/site_properties/"
    disabled_endpoints = ["update"]


class BuildingPropertyResources(BaseResources):
    endpoint_base_uri = "/building_properties/"
    disabled_endpoints = ["update"]


class StoreyPropertyResources(BaseResources):
    endpoint_base_uri = "/storey_properties/"
    disabled_endpoints = ["update"]


class SpacePropertyResources(BaseResources):
    endpoint_base_uri = "/space_properties/"
    disabled_endpoints = ["update"]


class ZonePropertyResources(BaseResources):
    endpoint_base_uri = "/zone_properties/"
    disabled_endpoints = ["update"]


class SitePropertyDataResources(BaseResources):
    endpoint_base_uri = "/site_property_data/"


class BuildingPropertyDataResources(BaseResources):
    endpoint_base_uri = "/building_property_data/"


class StoreyPropertyDataResources(BaseResources):
    endpoint_base_uri = "/storey_property_data/"


class SpacePropertyDataResources(BaseResources):
    endpoint_base_uri = "/space_property_data/"


class ZonePropertyDataResources(BaseResources):
    endpoint_base_uri = "/zone_property_data/"



"""BEMServer API client enums"""
import enum


class DataFormat(enum.Enum):
    csv = "application/csv"
    json = "application/json"


class Aggregation(enum.Enum):
    avg = "avg"
    sum = "sum"
    min = "min"
    max = "max"
    count = "count"


class BucketWidthUnit(enum.Enum):
    second = "second"
    minute = "minute"
    hour = "hour"
    day = "day"
    week = "week"
    month = "month"
    year = "year"


class EventLevel(enum.Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class StructuralElement(enum.Enum):
    site = "site"
    building = "building"
    storey = "storey"
    space = "space"
    zone = "zone"



"""BEMServer API client resources

/event_categories/ endpoints
/event_categories_by_users/ endpoints
/events/ endpoints
/events_by_sites/ endpoints
/events_by_buildings/ endpoints
/events_by_storeys/ endpoints
/events_by_spaces/ endpoints
/events_by_zones/ endpoints
"""
from .base import BaseResources


class EventCategoryResources(BaseResources):
    endpoint_base_uri = "/event_categories/"


class EventCategoryByUserResources(BaseResources):
    endpoint_base_uri = "/event_categories_by_users/"


class EventResources(BaseResources):
    endpoint_base_uri = "/events/"


class EventBySiteResources(BaseResources):
    endpoint_base_uri = "/events_by_sites/"
    disabled_endpoints = ["update"]


class EventByBuildingResources(BaseResources):
    endpoint_base_uri = "/events_by_buildings/"
    disabled_endpoints = ["update"]


class EventByStoreyResources(BaseResources):
    endpoint_base_uri = "/events_by_storeys/"
    disabled_endpoints = ["update"]


class EventBySpaceResources(BaseResources):
    endpoint_base_uri = "/events_by_spaces/"
    disabled_endpoints = ["update"]


class EventByZoneResources(BaseResources):
    endpoint_base_uri = "/events_by_zones/"
    disabled_endpoints = ["update"]



"""BEMServer API client errors"""


class BEMServerAPIVersionError(Exception):
    """BEMServer API version error"""


class BEMServerAPIError(Exception):
    """BEMServer API error"""

    def __init__(self, status_code=501):
        self.status_code = status_code


class BEMServerAPIValidationError(BEMServerAPIError):
    """BEMServer API validation error"""

    def __init__(self, errors=None):
        super().__init__(status_code=422)
        self.errors = errors


class BEMServerAPINotFoundError(BEMServerAPIError):
    """BEMServer API not found error"""

    def __init__(self):
        super().__init__(status_code=404)


class BEMServerAPINotModified(BEMServerAPIError):
    """BEMServer API not modified"""

    def __init__(self):
        super().__init__(status_code=304)


class BEMServerAPIAuthenticationError(BEMServerAPIError):
    """BEMServer API authentication error"""


class BEMServerAPIPreconditionError(BEMServerAPIError):
    """BEMServer API precondition error"""


class BEMServerAPIInternalError(BEMServerAPIError):
    """BEMServer API internal error"""

    def __init__(self, status_code=500):
        super().__init__(status_code=status_code)


class BEMServerAPIClientValueError(ValueError):
    """BEMServer API client value error"""



"""BEMServer API client resources base class"""


class BaseResources:
    endpoint_base_uri = ""
    disabled_endpoints = []

    def __init__(self, request_manager):
        self._req = request_manager

    def _verify_disabled(self, endpoint_name):
        if not hasattr(self, endpoint_name):
            raise ValueError(
                f"{self.__class__.__name__}.{endpoint_name} does not exist!"
            )
        if endpoint_name in self.disabled_endpoints:
            raise NotImplementedError(
                f"{self.__class__.__name__}.{endpoint_name} is disabled!"
            )

    def enpoint_uri_by_id(self, id):
        return f"{self.endpoint_base_uri}{str(id)}"

    def getall(self, *, etag=None, **kwargs):
        self._verify_disabled("getall")
        return self._req.getall(self.endpoint_base_uri, etag=etag, params=kwargs)

    def getone(self, id, *, etag=None):
        self._verify_disabled("getone")
        return self._req.getone(self.enpoint_uri_by_id(id), etag=etag)

    def create(self, payload):
        self._verify_disabled("create")
        return self._req.create(self.endpoint_base_uri, payload)

    def update(self, id, payload, *, etag=None):
        self._verify_disabled("update")
        return self._req.update(self.enpoint_uri_by_id(id), payload, etag=etag)

    def delete(self, id, *, etag=None):
        self._verify_disabled("delete")
        return self._req.delete(self.enpoint_uri_by_id(id), etag=etag)



"""BEMServer API client resources

/notifications/ endpoints
"""
from .base import BaseResources


class NotificationResources(BaseResources):
    endpoint_base_uri = "/notifications/"

    def count_by_campaign(self, *, etag=None, **kwargs):
        endpoint = f"{self.endpoint_base_uri}count_by_campaign"
        return self._req.getall(endpoint, etag=etag, params=kwargs)

    def mark_all_as_read(self, **kwargs):
        endpoint = f"{self.endpoint_base_uri}mark_all_as_read"
        return self._req._execute("PUT", endpoint, params=kwargs)



#!/usr/bin/env python3
"""BEMServer API client"""

from setuptools import setup, find_packages


# Get the long description from the README file
with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="bemserver-api-client",
    version="0.11.1",
    description="BEMServer API client",
    long_description=long_description,
    url="https://github.com/BEMServer/bemserver-api-client",
    author="Nobatek/INEF4",
    author_email="dfrederique@nobatek.inef4.com",
    license="AGPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        (
            "License :: OSI Approved :: "
            "GNU Affero General Public License v3 or later (AGPLv3+)"
        ),
    ],
    python_requires=">=3.9",
    install_requires=[
        "requests>=2.28.2",
        "packaging>=23.0",
    ],
    packages=find_packages(exclude=["tests*"]),
)



"""BEMServer API client response"""
import logging
import json

from .exceptions import (
    BEMServerAPIValidationError,
    BEMServerAPINotFoundError,
    BEMServerAPINotModified,
    BEMServerAPIAuthenticationError,
    BEMServerAPIPreconditionError,
    BEMServerAPIInternalError,
)


class BEMServerApiClientResponse:
    """API client response"""

    def __init__(self, raw_response, *, logger=None):
        self._logger = logger or logging.getLogger(__name__)
        self._raw_response = raw_response
        content_type = self._raw_response.headers.get("Content-Type", "")
        self._mimetype = content_type.split("; ")[0]

        # Process redirection or error, if any.
        if self.status_code == 304:
            raise BEMServerAPINotModified
        elif self.status_code < 300 or self.status_code > 500:
            self._raw_response.raise_for_status()
        else:
            self._process_client_error()

    @property
    def status_code(self):
        return self._raw_response.status_code

    @property
    def etag(self):
        return self._raw_response.headers.get("ETag", "")

    @property
    def pagination(self):
        """Get pagination data, if any.

        Example:
            {"total": 4, "total_pages": 1, "first_page": 1, "last_page": 1, "page": 1}
        """
        return json.loads(self._raw_response.headers.get("X-Pagination", "{}"))

    @property
    def is_json(self):
        """Check if the mimetype indicates JSON data, either
        :mimetype:`application/json` or :mimetype:`application/*+json`.
        """
        return self._mimetype == "application/json" or (
            self._mimetype.startswith("application/")
            and self._mimetype.endswith("+json")
        )

    @property
    def is_csv(self):
        """Check if the mimetype indicates CSV data, either
        :mimetype:`application/csv` or :mimetype:`application/*+csv`.
        """
        return self._mimetype == "application/csv" or (
            self._mimetype.startswith("application/")
            and self._mimetype.endswith("+csv")
        )

    @property
    def data(self):
        if self.is_json:
            return self._raw_response.json()
        return self._raw_response.content

    def _process_client_error(self):
        self._logger.error(f"{self.status_code} {self._raw_response.url}")

        # Authentication error
        if self.status_code in (401, 403):
            raise BEMServerAPIAuthenticationError(status_code=self.status_code)

        # Precondition error (etag)
        elif self.status_code in (412, 428):
            raise BEMServerAPIPreconditionError(status_code=self.status_code)

        # Resource not found
        elif self.status_code == 404:
            raise BEMServerAPINotFoundError

        # Conflict or validation error
        elif self.status_code in (409, 422):
            # TODO: rework this part
            errors = {}
            if self.is_json:
                if self.status_code == 409:
                    # Unique constraint error
                    if self.data.get("errors", {}).get("type") == "unique_constraint":
                        errors = {
                            # TODO: manage multiple columns constraint
                            field: ["Must be unique."]
                            for field in self.data["errors"]["fields"]
                        }
                    # Foreign key constraint error (and default case)
                    else:
                        errors = {"_general": ["Operation failed (409)."]}
                elif self.status_code == 422:
                    if "errors" in self.data:
                        # Marshmallow ValidationError
                        for loc in ("json", "query", "files"):
                            if loc in self.data["errors"]:
                                errors = {**errors, **self.data["errors"][loc]}
                                if "_schema" in errors:
                                    errors["_general"] = errors.pop("_schema")
                    # BEMServer ValidationError
                    elif "message" in self.data:
                        errors = {"_general": [self.data["message"]]}
            raise BEMServerAPIValidationError(errors=errors)

        # Issue in BEMServer
        raise BEMServerAPIInternalError(status_code=self.status_code)

    def toJSON(self):
        # Allows to set this response instance in a serializable object.
        return {
            "status_code": self.status_code,
            "data": self.data,
            "etag": self.etag,
            "pagination": self.pagination,
        }



"""BEMServer API client resources

/campaigns/ endpoints
/user_groups_by_campaigns/ endpoints
/campaign_scopes/ endpoints
/user_groups_by_campaign_scopes/ endpoints
"""
from .base import BaseResources


class CampaignResources(BaseResources):
    endpoint_base_uri = "/campaigns/"


class UserGroupByCampaignResources(BaseResources):
    endpoint_base_uri = "/user_groups_by_campaigns/"
    disabled_endpoints = ["update"]


class CampaignScopeResources(BaseResources):
    endpoint_base_uri = "/campaign_scopes/"


class UserGroupByCampaignScopeResources(BaseResources):
    endpoint_base_uri = "/user_groups_by_campaign_scopes/"
    disabled_endpoints = ["update"]
